#ifndef __SOUND_MISC_H_
#define __SOUND_MISC_H_

//////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2008 Targem Ltd. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

//
namespace m3d {
namespace sound {

	/// sound name typedef
	typedef		FixedString< 32, ClampInsertPolicy >	CueName_t;


	/// sound instance (channel) handle
	typedef				i32					CueId;
	const				CueId				InvalidCueId	=	-1;


	/// these flags specified when triggering sound playback
	enum PlaybackFlags
	{
		/// "simple" sounds (most UI and stereo sounds), no panning or 3d distance attenuation applied
		PF_NONE			=	0,
		/// panning calculation applied
		PF_POSITION		=	1,
		/// 3d distance attenuation applied
		PF_DISTANCE		=	2,

		/// normal 3d sounds
		PF_NORMAL3D		=	PF_POSITION | PF_DISTANCE,
	};


	/// return true if sound loaded
	bool				CheckCueExists( const char* cueName );
	/// return unique sound id (use it instead of string names to compare two sounds)
	u32					GetUniqueCueId( const char* cueName );

	/// start playback (cues are garbage-collected, you don't need to destroy it manually)
	/// FIXME: [Plus] take velocity into account for Doppler shift calculations
	CueId				PlayCue( const char* cueName, u32 flags, f32 volume = 1.0f, const Vec3& origin = ZeroVector, const Vec3& velocity = ZeroVector );
	CueId				PlayCueByUniqueId( u32 uniqueCueId, u32 flags, f32 volume = 1.0f, const Vec3& origin = ZeroVector, const Vec3& velocity = ZeroVector );

	/// cue instance control (but feel free to use SoundCueRef, its more handy)
	bool				Cue_IsValid( CueId cueId );
	bool				Cue_IsPlaying( CueId cueId );
	bool				Cue_IsStopped( CueId cueId );
	bool				Cue_IsPaused( CueId cueId );
	bool				Cue_IsPreparing( CueId cueId );
	bool				Cue_IsActive( CueId cueId );			//!< "active" cue means playing or preparing one
	f32					Cue_GetVolume( CueId cueId );
	void				Cue_SetVolume( CueId cueId, f32 volume );
	void				Cue_SetOrigin( CueId cueId, const Vec3& origin );
	f32					Cue_GetPitch( CueId cueId );
	void				Cue_SetPitch( CueId cueId, f32 pitch );
	void				Cue_Stop( CueId cueId );
	void				Cue_Pause( CueId cueId, bool pause );
	void				Cue_GetInfo( CueId cueId, OUT i32& positionMs, OUT i32& lengthMs );
	u32					Cue_GetFlags( CueId cueId );


	/// simple wrapper around CueId
	struct	SoundCueRef
	{
	public:
		/// ctors
						SoundCueRef() : m_cueId( InvalidCueId )							{}
		explicit		SoundCueRef( CueId cueId ) : m_cueId( cueId )					{}
		explicit		SoundCueRef( const SoundCueRef& rhs ) : m_cueId( rhs.m_cueId )	{}

		///
		CueId			GetCueId() const				{	return m_cueId;									}

		///
		SoundCueRef&	operator =( CueId rhs )			{	m_cueId = rhs; return *this;					}
		SoundCueRef&	operator =( SoundCueRef& rhs )	{	m_cueId = rhs.m_cueId; return *this;			}

		///
		bool			IsValid() const					{	return Cue_IsValid( m_cueId );					}
		bool			IsPlaying() const				{	return Cue_IsPlaying( m_cueId );				}
		bool			IsPaused() const				{	return Cue_IsPaused( m_cueId );					}
		bool			IsPreparing() const				{	return Cue_IsPreparing( m_cueId );				}
		bool			IsActive() const				{	return Cue_IsActive( m_cueId );					}
		bool			IsStopped() const				{	return Cue_IsStopped( m_cueId );				}

		///
		f32				GetVolume() const				{	return Cue_GetVolume( m_cueId );				}
		f32				GetPitch() const				{	return Cue_GetPitch( m_cueId );					}
		void			SetVolume( f32 volume )			{	Cue_SetVolume( m_cueId, volume );				}
		void			SetPitch( f32 pitch )			{	Cue_SetPitch( m_cueId, pitch );					}
		void			SetOrigin( const Vec3& org )	{	Cue_SetOrigin( m_cueId, org );					}

		///
		void			Pause( bool pause )				{	Cue_Pause( m_cueId, pause );					}
		void			Stop()							{	Cue_Stop( m_cueId ); m_cueId = InvalidCueId;	}
		void			GetInfo( OUT i32& positionMs, OUT i32& lengthMs ) const					
														{	Cue_GetInfo( m_cueId, positionMs, lengthMs );	}

	private:
		//
		CueId			m_cueId;
	};


	// make sure there are fully compatible
	M3D_STATIC_CHECK( sizeof( SoundCueRef ) == sizeof( CueId ) );


	/**
		WARNING: SoundEmitter does NOT stop its sounds when being removed, 
		you should do it explicitly via SoundEmitter::StopAll().

		Here's the example to justify this decision: when monster dies she
		plays her death sound and removes herself. If we stop sounds here, then
		death sound will be stopped immediately right after it has been started.
	*/
	class	SoundEmitter
	{
	public:
		///
		static			void					InitSystem();
		static			void					ShutdownSystem();

		/// factory
		static			SoundEmitter*			AllocateSoundEmitter();
		static			void					FreeSoundEmitter( SoundEmitter* e );

		/// ctor/dtor
						SoundEmitter();
						~SoundEmitter();

		/// should be called whenever audio source origin changes
		void			Update( const Vec3& origin, const Vec3& velocity );

		/// playback control
		CueId			StartSound( i32 channelId, const char* cueName, f32 volume = 1.0f, u32 flags = PF_NORMAL3D );
		void			StopSound( i32 channelId );
		void			StopSoundOnChannelRange( i32 channelIdMin, i32 channelIdMax );
		void			FadeSound( i32 channelId, f32 to, f32 over );
		bool			IsPlaying( i32 channelId ) const;
		void			PauseSound( i32 channelId, bool pause );

		// return initial flags of cue in specified channel
		u32				GetFlags( i32 channelId ) const;

		/// volume/pitch control
		void			SetVolume( i32 channelId, f32 vol );
		f32				GetVolume( i32 channelId );
		void			SetPitch( i32 channelId, f32 pitch );
		f32				GetPitch( i32 channelId );

		///
		void			GetInfo( i32 channelId, OUT i32& positionMs, OUT i32& lengthMs );

		/// stop all sounds
		void			StopAll();

		/// no place for new sound
		bool			IsFull() const;

	private:
		//
		Vec3			m_origin;

		//
		static const i32 MAX_SOUNDS_PER_EMITTER = 12;
		struct SoundDesc_t
		{
			CueId		cueId;
			i32			channelId;
		};
		FixedArray< SoundDesc_t, MAX_SOUNDS_PER_EMITTER >	m_sounds;
	};


	/// editor helpers
	i32					Editor_GetNumProjects();
	const char*			Editor_GetProjectName( i32 projectIdx );
	i32					Editor_GetNumCues( i32 projectIdx );
	const char*			Editor_GetCueName( i32 projectIdx, i32 cueIdx );

}	// namespace sound
}	// namespace m3d


#endif	// __SOUND_MISC_H_
