
#include "Precompiled.h"

#include "Engine/UI/DrawContext.h"
#include "TextureList.h"


using namespace m3d;
using namespace m3d::devgui;


//
bool TextureList::LoadTexture( const char* name, const char* fileName )
{
	if( Find( name ) )
	{
		M3D_LOG_ERR( "TextureList::LoadTexture: already loaded texture '%s'", name );
		return false;
	}

	Add( name, fileName, UiRect(), UiRect() );
	return true;
}


//
bool TextureList::LoadAtlas( const char* fileName )
{
	Parser parser;
	if( !parser.LoadFromFile( fileName ) )
	{
		M3D_LOG_WRN( "Can't load texture atlas %s", fileName );
		return false;
	}

	//
	StrTokenizer<> tok( " " );
			
	//
	Token token;
	while( parser.ReadToken( token ) )
	{
		ttl::String name;
		FilePath atlasTexture;
		UiRect r;
		UiRect scale9Grid;

		if( token != "Texture" )
		{
			M3D_LOG_ERR( "TextureList::LoadAtlas: 'Texture' expected" );
			return false;
		}

		if( !parser.ReadToken( token ) || token != "{" )
		{
			M3D_LOG_ERR( "TextureList::LoadAtlas: 'Texture' expected" );
			return false;
		}

		while( parser.ReadToken( token ) && token != "}" )
		{
			if( token == "name" )
			{
				parser.ReadToken( token );
				name.Set( token.Begin(), token.GetLength() );
			}
			else if( token == "fileName" )
			{
				parser.ReadToken( token );
				atlasTexture.assign( token.Begin(), token.End() );
			}
			else if( token == "rect" )
			{
				parser.ReadToken( token );
				tok.Parse( token.Begin(), token.GetLength() );
				if( tok.Argc() != 4 )
				{
					M3D_LOG_INF( "TextureList::LoadAtlas: bad rect" );
					return false;
				}

				r.x = Strtof( tok.Argv( 0 ) );
				r.y = Strtof( tok.Argv( 1 ) );
				r.w = Strtof( tok.Argv( 2 ) );
				r.h = Strtof( tok.Argv( 3 ) );
			}
			else if( token == "scale9grid" )
			{
				parser.ReadToken( token );
				tok.Parse( token.Begin(), token.GetLength() );
				if( tok.Argc() != 4 )
				{
					M3D_LOG_INF( "TextureList::LoadAtlas: bad scale9grid rect" );
					return false;
				}

				scale9Grid.x = Strtof( tok.Argv( 0 ) );
				scale9Grid.y = Strtof( tok.Argv( 1 ) );
				scale9Grid.w = Strtof( tok.Argv( 2 ) );
				scale9Grid.h = Strtof( tok.Argv( 3 ) );
			}
		}

		/*M3D_LOG_INF( "tex: %s, atlas: %s, rect = { %.1f, %.1f, %.1f, %.1f }", 
			name.c_str(), atlasTexture.c_str(), r.x, r.y, r.w, r.h );*/

		//
		if( Find( name.c_str() ) )
		{
			M3D_LOG_ERR( "TextureList::LoadAtlas: duplicate atlas texture '%s'", name.c_str() );
			//return false;
		}
		else
		{
			if( !scale9Grid.Empty() )
			{
				// make sure slicing rect is inside the main rect area
				UiRect slice = scale9Grid;
				slice.Offset( r.TopLeft() );
				if( slice.x < r.x || slice.y < r.y || slice.Right() > r.Right() || slice.Bottom() > r.Bottom() )
				{
					M3D_LOG_ERR( "TextureList::LoadAtlas: 'scale9Grid' for texture '%s' is outside of texture bounds", name.c_str() );
					scale9Grid.Zero();
				}
			}

			Add( name.c_str(), atlasTexture.c_str(), r, scale9Grid );
		}
	}

	return true;
}


//
Texture_t* TextureList::Find( const char* textureName )
{
	for( List_t::iterator it = m_list.begin(); it != m_list.end(); ++it )
	{
		Texture_t& t = *it;
		if( StrIEq( t.name.c_str(), textureName ) )
			return &t;
	}

	return 0;
}


//
void TextureList::Free()
{
	for( List_t::iterator it = m_list.begin(); it != m_list.end(); ++it )
	{
		Texture_t& t = *it;
		M3D_SAFE_RELEASE( t.tex );
	}

	m_list.clear();
}


//
void TextureList::Add( const char* name, const char* atlasTexture, const UiRect& baseUvbounds, const UiRect& scale9Grid )
{
	bool slicing = !scale9Grid.Empty();

	//
	u32 numTextures = slicing ? 9 : 1;
	if( m_list.size() > ( MAX_TEXTURES - numTextures ) )
	{
		M3D_LOG_ERR( "Too many textures" );
		return;
	}

	//
	rend::Texture* atlas = rend::AddFileTexture( atlasTexture, true );
	u32 atlasw, atlash;
	atlas->GetDims( atlasw, atlash );
	//
	UiRect uvbounds = baseUvbounds;
	if( uvbounds.Empty() )
		uvbounds = UiRect( 0, 0, ( f32 )atlasw, ( f32 )atlash );

	//
	if( slicing )
	{
		UiRect r = uvbounds;
		UiRect i = scale9Grid;
		UiRect grid[ 3 * 3 ] = {
			UiRect( 0.0f,		0.0f,			i.x,				i.y ),
			UiRect( i.x,		0.0f,			i.w,				i.y ),
			UiRect( i.Right(),	0.0f,			r.w - i.Right(),	i.y ),

			UiRect( 0.0f,		i.y,			i.x,				i.h ),
			UiRect( i.x,		i.y,			i.w,				i.h ),
			UiRect( i.Right(),	i.y,			r.w - i.Right(),	i.h ),

			UiRect( 0.0f,		i.Bottom(),		i.x,				r.h - i.Bottom() ),
			UiRect( i.x,		i.Bottom(),		i.w,				r.h - i.Bottom() ),
			UiRect( i.Right(),	i.Bottom(),		r.w - i.Right(),	r.h - i.Bottom() ),
		};

		//
		for( i32 ii = 0; ii < 3 * 3; ++ii )
		{
			m_list.push_back( Texture_t() );
			Texture_t& t = m_list.back();
			t.name = StrFormat( "%s_%d", name, ii + 1 );
			t.tex = atlas;
			t.tex->AddRef();
			
			UiRect slice = grid[ ii ];
			slice.Offset( r.TopLeft() );
		
			t.w = slice.w;
			t.h = slice.h;
			t.u0 = slice.x / atlasw;
			t.v0 = slice.y / atlash;
			t.u1 = slice.Right() / atlasw;
			t.v1 = slice.Bottom() / atlash;
		}
	}
	else
	{
		m_list.push_back( Texture_t() );
		Texture_t& t = m_list.back();
		t.name = name;
		t.tex = atlas;
		t.tex->AddRef();

		//
		t.w = uvbounds.w;
		t.h = uvbounds.h;
		t.u0 = uvbounds.x / atlasw;
		t.v0 = uvbounds.y / atlash;
		t.u1 = uvbounds.Right() / atlasw;
		t.v1 = uvbounds.Bottom() / atlash;
	}
}
